try:
	from urlparse import urljoin
except ImportError:
	# Python 3
	from urllib.parse import urljoin
import time
import logging

import requests
import univention.admin.uldap as udm_uldap
import univention.testing.ucr
from univention.ox.wait_for_ox_server import wait_for_ox_server


OX_ATTR_TO_LDAP_ATTR = dict([
	('display_name', 'oxDisplayName'),
	('email1', 'mailPrimaryAddress'),
	('first_name', 'givenName'),
	('login_info', 'uid'),
	('timezone', 'oxTimeZone')
])
LDAP_ATTR_TO_OX_ATTR = dict((v, k) for k, v in OX_ATTR_TO_LDAP_ATTR.items())


class OxHttpApiError(Exception):
	pass


class NotFoundError(OxHttpApiError):
	pass


class LoginError(OxHttpApiError):
	pass


class AttributeSyncError(OxHttpApiError):
	pass


class ApiError(OxHttpApiError):
	pass


class UnkownMethodError(OxHttpApiError):
	pass


class OXHttpApi(object):
	def __init__(self, name, password):
		"""
		Log user in.

		:param name: str: username for authentication
		:param password: str: password for authentication
		"""
		assert isinstance(name, basestring)
		assert isinstance(password, basestring)
		self.username = name  # 'test1'
		self.password = password  # 'univention'
		self.session = None
		self.cookies = None
		self._lo = None
		with univention.testing.ucr.UCSTestConfigRegistry() as ucr:
			self.base_url = 'http://{}.{}/ajax/'.format(ucr['hostname'], ucr['domainname'])
		self.logger = logging.getLogger('ucs-test')
		self.logger.setLevel(logging.DEBUG)
		my_handler_name = 'ox_http_api_client'
		if not any(handler.get_name() == my_handler_name for handler in self.logger.handlers):
			handler = logging.StreamHandler()
			handler.set_name(my_handler_name)
			handler.setFormatter(logging.Formatter(
				fmt='%(asctime)s %(levelname)s: %(message)s',
				datefmt='%Y-%m-%d %H:%M:%S'))
			handler.setLevel(logging.DEBUG)
			self.logger.addHandler(handler)

		wait_for_ox_server()
		try:
			self._login()
		except LoginError as exc:
			if 'Please try again later' in str(exc):
				self.logger.error('Error logging in. Sleeping 10s and retrying once. Error was:\n%s', str(exc))
				time.sleep(10)
				self._login()
		self.logger.info('Logged in as user "{}".'.format(self.username))

	@staticmethod
	def init_admin_credentials():
		"""
		Get initiated OXHttpApi instance.

		:return: OXHttpApi instance using oxadmin credentials from disk.
		"""
		with univention.testing.ucr.UCSTestConfigRegistry() as ucr:
			context_id = ucr['ox/context/id']
		admin_user = 'oxadmin'
		admin_password = open('/etc/ox-secrets/context{}.secret'.format(context_id), 'rb').read().strip()
		return OXHttpApi(admin_user, admin_password)

	@staticmethod
	def _get_json(obj):
		"""
		Handle different requests versions.

		:param obj: response object
		:return: dict: dict from parsed json
		"""
		if callable(obj.json):
			return obj.json()
		else:
			return obj.json

	def _login(self):
		url = urljoin(self.base_url, 'login?action=login')
		data = dict(name=self.username, password=self.password)
		response = requests.post(url, data=data)
		if response.status_code != 200:
			raise LoginError('Received status_code={} when trying to log in: {}'.format(
				response.status_code,
				response.content))
		elif 'error' in self._get_json(response):
			raise LoginError('Error logging in: {}'.format(self._get_json(response)['error']))
		else:
			self.session = self._get_json(response)['session']
			self.cookies = response.cookies

	def _call_api(self, method, url_end, data=None, **params):
		url = urljoin(self.base_url, url_end)
		params.update(dict(session=self.session))
		kwargs = dict(url=url, params=params, cookies=self.cookies)
		if data:
			kwargs['data'] = data
		meth = getattr(requests, method)
		response = meth(**kwargs)
		json_response = self._get_json(response)
		try:
			return json_response['data']
		except KeyError:
			if 'error_desc' in json_response or 'error' in json_response:
				raise ApiError('When calling {} "{}" with params="{}" and data={}: {}'.format(
					method.upper(), url, params, data, json_response.get('error_desc', json_response['error'])))
			else:
				return json_response

	def get(self, url_end, **params):
		return self._call_api('get', url_end, **params)

	def post(self, url_end, data, **params):
		return self._call_api('post', url_end, data, **params)

	def put(self, url_end, data=None, **params):
		return self._call_api('put', url_end, data, **params)

	def get_context_id(self):
		"""
		Get the context ID of the logged in user.

		:return: dict: {u'data': 10}
		"""
		return self.get('config/context_id')

	def get_user(self, oid=None):
		"""

		:param oid: int: Object ID of the requested user, current user if None
		:return: dict
		"""
		kwargs = dict(action='get')
		if oid:
			kwargs['id'] = oid
		return self.user(**kwargs)

	def list_users(self, columns=None):
		"""
		To find IDs of columns look at:
		- http://oxpedia.org/wiki/index.php?title=HTTP_API#CommonObjectData
		- http://oxpedia.org/wiki/index.php?title=HTTP_API#DetailedContactData
		- http://oxpedia.org/wiki/index.php?title=HTTP_API#DetailedUserData

		:param columns: list: list of ints (IDs of columns to show)
		:return: list of lists
		"""
		if columns:
			assert isinstance(columns, list)
		else:
			columns = [1, 4, 223, 500, 501, 502, 524]
		return self.user('all', columns=",".join(str(x) for x in columns))

	def export(self, action, folder, columns=None, export_dlists=None):
		"""

		:param action: str: one of 'CSV', 'ICAL', 'VCARD'
		:param folder: int/str: ObjectID of the calendar/contact folder whose contents
		are to be exported.
		:param columns: list: optional, only for CSV
		:param export_dlists: bool: optional, only for CSV
		:return: An InputStream containing the file, of the MIME type
		text/csv | text/calendar | text/x-vcard
		"""
		assert action in ('CSV', 'ICAL', 'VCARD')
		return self.get('export?action={}'.format(action), folder=folder)

	def folders(self, action, data=None, **params):
		"""
		Access/change OX drive folder information.

		List all folders with IDs: folders('root', columns='1,300')

		:param action: str: GET: 'root', 'list', 'path', 'updates', 'get', 'allVisible', 'shares'
		PUT: 'update', 'new', 'delete', 'clear', 'notify'
		:param data: dict/bytes/file: dict, bytes, or file-like object to send in the body of a PUT request
		:param params: parameters to pass to OX API
		:return: list of lists
		"""
		if action in ('root', 'list', 'path', 'updates', 'get', 'allVisible', 'shares'):
			method = self.get
		elif action in ('update', 'new', 'delete', 'clear', 'notify'):
			method = self.put
		else:
			raise UnkownMethodError('Unknown action "{}" for contacts.'.format(action))
		return method('folders?action={}'.format(action), data=data, **params)

	def contacts(self, action, data=None, **params):
		"""
		Access/change contact information.

		Ex: list all global contacts: contacts_get('all', folder=6, columns='1,20,223,500,501,502')

		:param action: str: GET: 'all', 'updates', 'get', 'getuser', 'anniversaries', 'birthdays', 'autocomplete'
		PUT: 'list', 'listuser', 'update', 'new', 'delete', 'search', 'advancedSearch'
		:param data: dict/bytes/file: dict, bytes, or file-like object to send in the body of a PUT request
		:param params: parameters to pass to OX API
		:return: list of lists
		"""
		if action in ('all', 'updates', 'get', 'getuser', 'anniversaries', 'birthdays', 'autocomplete'):
			method = self.get
		elif action in ('list', 'listuser', 'update', 'new', 'delete', 'search', 'advancedSearch'):
			method = self.put
		else:
			raise UnkownMethodError('Unknown action "{}" for contacts.'.format(action))
		return method('contacts?action={}'.format(action), data=data, **params)

	def user(self, action, data=None, **params):
		"""
		Access/change user information.

		:param action: str: GET: ''all', 'get', 'getAttribute'
		PUT: 'list', 'update', 'search', 'setAttribute'
		:param data: dict/bytes/file: dict, bytes, or file-like object to send in the body of a PUT request
		:param params: parameters to pass to OX API
		:return: list of lists
		"""
		if action in ('all', 'get', 'getAttribute'):
			method = self.get
		elif action in ('list', 'update', 'search', 'setAttribute'):
			method = self.put
		else:
			raise UnkownMethodError('Unknown action "{}" for contacts.'.format(action))
		return method('user?action={}'.format(action), data=data, **params)

	def username_to_id(self, username):
		"""
		Get the user ID (not object ID) of a user.

		:param username: str: surname
		:return: int: ID or NotFoundError
		"""
		assert isinstance(username, basestring)
		user_list = self.put('user?action=all', data=bytes(username), columns="1,615")
		for oid, _username, _none in user_list:
			if _username == username:
				return oid
		raise NotFoundError('Could not get object ID of user "{}".'.format(username))

	@property
	def lo(self):
		if not self._lo:
			self._lo = udm_uldap.getMachineConnection(ldap_master=True)[0]
		return self._lo

	def compare_ldap_and_ox_user(self, username):
		oid = self.username_to_id(username)
		self.logger.info('User "{}" has object ID "{}".'.format(username, oid))

		ox_user = self.get_user(oid)
		_dn, ldap_user = self.lo.search(filter='(&(objectClass=posixAccount)(uid={}))'.format(username))[0]
		for ox_attr, ldap_attr in OX_ATTR_TO_LDAP_ATTR.items():
			if ox_user[ox_attr] != ldap_user[ldap_attr][0]:
				raise AttributeSyncError('Attributes not equal: ox_user[{}]={} ldap_user[][0]={}'.format(
					ox_attr, ox_user[ox_attr], ldap_attr, ldap_user[ldap_attr][0]))
		self.logger.info('OK: Attributes in LDAP and in OX-DB synced correctly.')
